import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat

files = [("test_basic_v7.mat", "v7"), ("test_basic_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadStructCell:
    """Tests for reading Struct and Cell arrays from MAT-files"""

    def test_cell_scalar(self, filename, version):
        """Test reading Cell Scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_scalar"])

        cell_scalar = np.array(
            [[np.array(["text"], dtype=np.str_).reshape(1, 1)]], dtype=object
        ).reshape(1, 1)

        np.testing.assert_array_equal(mdict["cell_scalar"], cell_scalar, strict=True)

    def test_cell_array(self, filename, version):
        """Test reading Cell Array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_array"])

        cell1 = np.array(["A"], dtype=np.str_)
        cell2 = np.array([[1, 2], [3, 4]], dtype=np.float64).reshape(2, 2)
        cell3 = np.array(
            [[np.array([[True]])], [np.array([[False]])]], dtype=object
        ).reshape(1, 2)

        assert mdict["cell_array"].shape == (1, 3)
        assert mdict["cell_array"].dtype == object

        np.testing.assert_array_equal(mdict["cell_array"][0, 0], cell1, strict=True)
        np.testing.assert_array_equal(mdict["cell_array"][0, 1], cell2, strict=True)
        np.testing.assert_array_equal(mdict["cell_array"][0, 2], cell3, strict=True)

    def test_cell_nested(self, filename, version):
        """Test reading Nested Cell Array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_nested"])

        level1 = np.array(["level1"], dtype=np.str_)
        level2 = np.array(["level2"], dtype=np.str_)
        level3a = np.array(["level3"], dtype=np.str_)
        level3b = np.array([[123]], dtype=np.float64).reshape(1, 1)

        # Root
        assert mdict["cell_nested"].shape == (1, 1)
        assert mdict["cell_nested"].dtype == object

        # Level 1
        assert mdict["cell_nested"][0, 0].shape == (1, 2)
        assert mdict["cell_nested"][0, 0].dtype == object
        np.testing.assert_array_equal(
            mdict["cell_nested"][0, 0][0, 0], level1, strict=True
        )
        assert mdict["cell_nested"][0, 0][0, 1].shape == (1, 1)
        assert mdict["cell_nested"][0, 0][0, 1].dtype == object

        # Level 2
        assert mdict["cell_nested"][0, 0][0, 1][0, 0].shape == (1, 2)
        assert mdict["cell_nested"][0, 0][0, 1][0, 0].dtype == object
        np.testing.assert_array_equal(
            mdict["cell_nested"][0, 0][0, 1][0, 0][0, 0], level2, strict=True
        )
        assert mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1].shape == (1, 1)
        assert mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1].dtype == object

        # Level 3
        assert mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].shape == (1, 2)
        assert mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].dtype == object
        np.testing.assert_array_equal(
            mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 0],
            level3a,
            strict=True,
        )
        np.testing.assert_array_equal(
            mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 1],
            level3b,
            strict=True,
        )

    def test_cell_empty(self, filename, version):
        """Test reading Empty Cell Array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_empty"])

        assert mdict["cell_empty"].shape == (0, 0)
        assert mdict["cell_empty"].dtype == object

    def test_struct_scalar(self, filename, version):
        """Test reading Struct Scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_scalar"])

        name = np.array(["test"], dtype=np.str_)
        value = np.array([[123]], dtype=np.float64).reshape(1, 1)
        data = np.array([[1, 2], [3, 4]], dtype=np.float64).reshape(2, 2)

        assert mdict["struct_scalar"].shape == (1, 1)
        assert set(mdict["struct_scalar"].dtype.names) == {"name", "value", "data"}

        np.testing.assert_array_equal(
            mdict["struct_scalar"]["name"][0, 0], name, strict=True
        )
        np.testing.assert_array_equal(
            mdict["struct_scalar"]["value"][0, 0], value, strict=True
        )
        np.testing.assert_array_equal(
            mdict["struct_scalar"]["data"][0, 0], data, strict=True
        )

    def test_struct_array(self, filename, version):
        """Test reading Struct Array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_array"])

        id1 = np.array([[1]], dtype=np.float64).reshape(1, 1)
        id2 = np.array([[2]], dtype=np.float64).reshape(1, 1)
        info1 = np.array(["first"], dtype=np.str_)
        info2 = np.array(["second"], dtype=np.str_)

        assert mdict["struct_array"].shape == (1, 2)
        assert set(mdict["struct_array"].dtype.names) == {"id", "info"}

        np.testing.assert_array_equal(
            mdict["struct_array"]["id"][0, 0], id1, strict=True
        )
        np.testing.assert_array_equal(
            mdict["struct_array"]["info"][0, 0], info1, strict=True
        )
        np.testing.assert_array_equal(
            mdict["struct_array"]["id"][0, 1], id2, strict=True
        )
        np.testing.assert_array_equal(
            mdict["struct_array"]["info"][0, 1], info2, strict=True
        )

    def test_struct_empty(self, filename, version):
        """Test reading Empty Struct Array from MAT-file"""
        from matio.utils.matclass import EmptyMatStruct

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_empty"])
        assert "struct_empty" in mdict
        assert isinstance(mdict["struct_empty"], EmptyMatStruct)
        assert mdict["struct_empty"].shape == (0, 0)

    def test_struct_no_fields(self, filename, version):
        """Test reading Empty Struct Array from MAT-file"""
        from matio.utils.matclass import EmptyMatStruct

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_no_fields"])
        assert "struct_no_fields" in mdict
        assert isinstance(mdict["struct_no_fields"], EmptyMatStruct)
        assert mdict["struct_no_fields"].shape == (1, 1)

    def test_struct_nested(self, filename, version):
        """Test reading Nested Struct Array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_nested"])

        assert mdict["struct_nested"].shape == (1, 1)
        assert set(mdict["struct_nested"].dtype.names) == {"level1"}

        level1 = mdict["struct_nested"]["level1"][0, 0]
        assert level1.shape == (1, 1)
        assert set(level1.dtype.names) == {"level2"}

        level2 = level1["level2"][0, 0]
        assert level2.shape == (1, 1)
        assert set(level2.dtype.names) == {"level3", "cell"}

        level3 = level2["level3"][0, 0]
        level3_np = np.array([[42]], dtype=np.float64).reshape(1, 1)
        assert level3.shape == (1, 1)
        assert set(level3.dtype.names) == {"value"}
        np.testing.assert_array_equal(level3["value"][0, 0], level3_np, strict=True)

        cell = level2["cell"][0, 0]
        assert cell.shape == (1, 1)
        assert cell.dtype == object

        assert cell[0, 0].shape == (1, 2)
        assert cell[0, 0].dtype == object

        cell1 = np.array(["nested"], dtype=np.str_)
        np.testing.assert_array_equal(cell[0, 0][0, 0], cell1, strict=True)

        assert cell[0, 0][0, 1].shape == (1, 1)
        assert set(cell[0, 0][0, 1].dtype.names) == {"a", "b"}

        cell2a = np.array([[1]], dtype=np.float64).reshape(1, 1)
        cell2b = np.array([[2]], dtype=np.float64).reshape(1, 1)
        np.testing.assert_array_equal(cell[0, 0][0, 1]["a"][0, 0], cell2a, strict=True)
        np.testing.assert_array_equal(cell[0, 0][0, 1]["b"][0, 0], cell2b, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveStructCell:

    def test_cell_scalar(self, filename, version):
        """Test writing cell scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_scalar"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cell_scalar"])

            assert mload["cell_scalar"].shape == mdict["cell_scalar"].shape
            assert mload["cell_scalar"].dtype == mdict["cell_scalar"].dtype
            np.testing.assert_array_equal(
                mdict["cell_scalar"][0, 0], mload["cell_scalar"][0, 0], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_cell_array(self, filename, version):
        """Test writing cell array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cell_array"])

            assert mload["cell_array"].shape == mdict["cell_array"].shape
            assert mload["cell_array"].dtype == mdict["cell_array"].dtype
            for i in np.ndindex(mdict["cell_array"].shape):
                np.testing.assert_array_equal(
                    mdict["cell_array"][i], mload["cell_array"][i], strict=True
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_cell_nested(self, filename, version):
        """Test writing nested cell array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_nested"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cell_nested"])

            # Root
            assert mload["cell_nested"].shape == mdict["cell_nested"].shape
            assert mload["cell_nested"].dtype == mdict["cell_nested"].dtype

            # Level 1
            assert mload["cell_nested"][0, 0].shape == mdict["cell_nested"][0, 0].shape
            assert mload["cell_nested"][0, 0].dtype == mdict["cell_nested"][0, 0].dtype
            np.testing.assert_array_equal(
                mdict["cell_nested"][0, 0][0, 0],
                mload["cell_nested"][0, 0][0, 0],
                strict=True,
            )
            assert (
                mload["cell_nested"][0, 0][0, 1].shape
                == mdict["cell_nested"][0, 0][0, 1].shape
            )
            assert (
                mload["cell_nested"][0, 0][0, 1].dtype
                == mdict["cell_nested"][0, 0][0, 1].dtype
            )

            # Level 2
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0].shape
                == mdict["cell_nested"][0, 0][0, 1][0, 0].shape
            )
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0].dtype
                == mdict["cell_nested"][0, 0][0, 1][0, 0].dtype
            )
            np.testing.assert_array_equal(
                mdict["cell_nested"][0, 0][0, 1][0, 0][0, 0],
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 0],
                strict=True,
            )
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1].shape
                == mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1].shape
            )
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1].dtype
                == mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1].dtype
            )

            # Level 3
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].shape
                == mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].shape
            )
            assert (
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].dtype
                == mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0].dtype
            )
            np.testing.assert_array_equal(
                mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 0],
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 0],
                strict=True,
            )
            np.testing.assert_array_equal(
                mdict["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 1],
                mload["cell_nested"][0, 0][0, 1][0, 0][0, 1][0, 0][0, 1],
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_cell_empty(self, filename, version):
        """Test writing cell empty to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cell_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cell_empty"])

            assert mload["cell_empty"].shape == mdict["cell_empty"].shape
            assert mload["cell_empty"].dtype == mdict["cell_empty"].dtype

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_struct_scalar(self, filename, version):
        """Test writing struct scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_scalar"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["struct_scalar"])

            assert mload["struct_scalar"].shape == mdict["struct_scalar"].shape
            assert set(mload["struct_scalar"].dtype.names) == set(
                mdict["struct_scalar"].dtype.names
            )
            for name in mload["struct_scalar"].dtype.names:
                np.testing.assert_array_equal(
                    mdict["struct_scalar"][name][0, 0],
                    mload["struct_scalar"][name][0, 0],
                    strict=True,
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_struct_array(self, filename, version):
        """Test writing struct array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["struct_array"])

            assert mload["struct_array"].shape == mdict["struct_array"].shape
            assert set(mload["struct_array"].dtype.names) == set(
                mdict["struct_array"].dtype.names
            )
            for name in mload["struct_array"].dtype.names:
                for i in np.ndindex(mdict["struct_array"].shape):
                    np.testing.assert_array_equal(
                        mdict["struct_array"][name][i],
                        mload["struct_array"][name][i],
                        strict=True,
                    )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_struct_empty(self, filename, version):
        """Test writing struct empty to MAT-file"""
        from matio.utils.matclass import EmptyMatStruct

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["struct_empty"])

            assert "struct_empty" in mload
            assert isinstance(mload["struct_empty"], EmptyMatStruct)
            assert mload["struct_empty"].shape == mdict["struct_empty"].shape

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_struct_no_fields(self, filename, version):
        """Test writing struct no fields to MAT-file"""
        from matio.utils.matclass import EmptyMatStruct

        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_no_fields"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["struct_no_fields"])

            assert "struct_no_fields" in mload
            assert isinstance(mload["struct_no_fields"], EmptyMatStruct)
            assert mload["struct_no_fields"].shape == mdict["struct_no_fields"].shape

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_struct_nested(self, filename, version):
        """Test writing struct nested to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["struct_nested"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["struct_nested"])

            # Root
            assert mload["struct_nested"].shape == mdict["struct_nested"].shape
            assert set(mload["struct_nested"].dtype.names) == set(
                mdict["struct_nested"].dtype.names
            )

            level1_md = mdict["struct_nested"]["level1"][0, 0]
            level1_ml = mload["struct_nested"]["level1"][0, 0]
            assert level1_ml.shape == level1_md.shape
            assert set(level1_ml.dtype.names) == set(level1_md.dtype.names)

            level2_md = level1_md["level2"][0, 0]
            level2_ml = level1_ml["level2"][0, 0]
            assert level2_ml.shape == level2_md.shape
            assert set(level2_ml.dtype.names) == set(level2_md.dtype.names)

            level3_md = level2_md["level3"][0, 0]
            level3_ml = level2_ml["level3"][0, 0]
            assert level3_ml.shape == level3_md.shape
            assert set(level3_ml.dtype.names) == set(level3_md.dtype.names)
            np.testing.assert_array_equal(
                level3_md["value"][0, 0], level3_ml["value"][0, 0], strict=True
            )

            cell_md = level2_md["cell"][0, 0]
            cell_ml = level2_ml["cell"][0, 0]
            assert cell_ml.shape == cell_md.shape
            assert cell_ml.dtype == cell_md.dtype

            assert cell_ml[0, 0].shape == cell_md[0, 0].shape
            assert cell_ml[0, 0].dtype == cell_md[0, 0].dtype
            np.testing.assert_array_equal(
                cell_md[0, 0][0, 0], cell_ml[0, 0][0, 0], strict=True
            )
            assert cell_ml[0, 0][0, 1].shape == cell_md[0, 0][0, 1].shape
            assert cell_ml[0, 0][0, 1].dtype == cell_md[0, 0][0, 1].dtype
            assert set(cell_ml[0, 0][0, 1].dtype.names) == set(
                cell_md[0, 0][0, 1].dtype.names
            )
            for name in cell_ml[0, 0][0, 1].dtype.names:
                np.testing.assert_array_equal(
                    cell_md[0, 0][0, 1][name][0, 0],
                    cell_ml[0, 0][0, 1][name][0, 0],
                    strict=True,
                )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



"""Constants and classes for matlab 5 read and write"""

# Copyright (c) 2001-2002 Enthought, Inc. 2003, SciPy Developers.
# All rights reserved.
#
# Modified by foreverallama (c) 2025
# https://github.com/foreverallama/matio
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following
#    disclaimer in the documentation and/or other materials provided
#    with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from enum import IntEnum

import numpy as np

__all__ = [
    "MDTYPES",
    "NP_TO_MTYPES",
    "NP_TO_MXTYPES",
    "codecs_template",
    "mclass_dtypes_template",
    "mclass_info",
    "mdtypes_template",
    "miTypes",
    "mxTypes",
]


class miTypes(IntEnum):
    miINT8 = 1
    miUINT8 = 2
    miINT16 = 3
    miUINT16 = 4
    miINT32 = 5
    miUINT32 = 6
    miSINGLE = 7
    miDOUBLE = 9
    miINT64 = 12
    miUINT64 = 13
    miMATRIX = 14
    miCOMPRESSED = 15
    miUTF8 = 16
    miUTF16 = 17
    miUTF32 = 18


class mxTypes(IntEnum):
    mxCELL_CLASS = 1
    mxSTRUCT_CLASS = 2
    mxOBJECT_CLASS = 3
    mxCHAR_CLASS = 4
    mxSPARSE_CLASS = 5
    mxDOUBLE_CLASS = 6
    mxSINGLE_CLASS = 7
    mxINT8_CLASS = 8
    mxUINT8_CLASS = 9
    mxINT16_CLASS = 10
    mxUINT16_CLASS = 11
    mxINT32_CLASS = 12
    mxUINT32_CLASS = 13
    mxINT64_CLASS = 14
    mxUINT64_CLASS = 15
    mxFUNCTION_CLASS = 16
    mxOPAQUE_CLASS = 17


mdtypes_template = {
    miTypes.miINT8: "i1",
    miTypes.miUINT8: "u1",
    miTypes.miINT16: "i2",
    miTypes.miUINT16: "u2",
    miTypes.miINT32: "i4",
    miTypes.miUINT32: "u4",
    miTypes.miSINGLE: "f4",
    miTypes.miDOUBLE: "f8",
    miTypes.miINT64: "i8",
    miTypes.miUINT64: "u8",
    miTypes.miUTF8: "u1",
    miTypes.miUTF16: "u2",
    miTypes.miUTF32: "u4",
    "file_header": [
        ("description", "S116"),
        ("subsystem_offset", "i8"),
        ("version", "u2"),
        ("endian_test", "S2"),
    ],
    "tag_full": [("mdtype", "u4"), ("byte_count", "u4")],
    "tag_smalldata": [("byte_count_mdtype", "u4"), ("data", "S4")],
    "array_flags": [
        ("data_type", "u4"),
        ("byte_count", "u4"),
        ("flags_class", "u4"),
        ("nzmax", "u4"),
    ],
    "U1": "U1",
}

mclass_dtypes_template = {
    mxTypes.mxINT8_CLASS: "i1",
    mxTypes.mxUINT8_CLASS: "u1",
    mxTypes.mxINT16_CLASS: "i2",
    mxTypes.mxUINT16_CLASS: "u2",
    mxTypes.mxINT32_CLASS: "i4",
    mxTypes.mxUINT32_CLASS: "u4",
    mxTypes.mxINT64_CLASS: "i8",
    mxTypes.mxUINT64_CLASS: "u8",
    mxTypes.mxSINGLE_CLASS: "f4",
    mxTypes.mxDOUBLE_CLASS: "f8",
}

mclass_info = {
    mxTypes.mxINT8_CLASS: "int8",
    mxTypes.mxUINT8_CLASS: "uint8",
    mxTypes.mxINT16_CLASS: "int16",
    mxTypes.mxUINT16_CLASS: "uint16",
    mxTypes.mxINT32_CLASS: "int32",
    mxTypes.mxUINT32_CLASS: "uint32",
    mxTypes.mxINT64_CLASS: "int64",
    mxTypes.mxUINT64_CLASS: "uint64",
    mxTypes.mxSINGLE_CLASS: "single",
    mxTypes.mxDOUBLE_CLASS: "double",
    mxTypes.mxCELL_CLASS: "cell",
    mxTypes.mxSTRUCT_CLASS: "struct",
    mxTypes.mxOBJECT_CLASS: "object",
    mxTypes.mxCHAR_CLASS: "char",
    mxTypes.mxSPARSE_CLASS: "sparse",
    mxTypes.mxFUNCTION_CLASS: "function_handle",
    mxTypes.mxOPAQUE_CLASS: "opaque",
}

NP_TO_MTYPES = {
    "f8": miTypes.miDOUBLE,
    "c32": miTypes.miDOUBLE,
    "c24": miTypes.miDOUBLE,
    "c16": miTypes.miDOUBLE,
    "f4": miTypes.miSINGLE,
    "c8": miTypes.miSINGLE,
    "i8": miTypes.miINT64,
    "i4": miTypes.miINT32,
    "i2": miTypes.miINT16,
    "i1": miTypes.miINT8,
    "u8": miTypes.miUINT64,
    "u4": miTypes.miUINT32,
    "u2": miTypes.miUINT16,
    "u1": miTypes.miUINT8,
    "S1": miTypes.miUINT8,
    "U1": miTypes.miUTF16,
    "b1": miTypes.miUINT8,
}


NP_TO_MXTYPES = {
    "f8": mxTypes.mxDOUBLE_CLASS,
    "c32": mxTypes.mxDOUBLE_CLASS,
    "c24": mxTypes.mxDOUBLE_CLASS,
    "c16": mxTypes.mxDOUBLE_CLASS,
    "f4": mxTypes.mxSINGLE_CLASS,
    "c8": mxTypes.mxSINGLE_CLASS,
    "i8": mxTypes.mxINT64_CLASS,
    "i4": mxTypes.mxINT32_CLASS,
    "i2": mxTypes.mxINT16_CLASS,
    "i1": mxTypes.mxINT8_CLASS,
    "u8": mxTypes.mxUINT64_CLASS,
    "u4": mxTypes.mxUINT32_CLASS,
    "u2": mxTypes.mxUINT16_CLASS,
    "u1": mxTypes.mxUINT8_CLASS,
    "S1": mxTypes.mxUINT8_CLASS,
    "b1": mxTypes.mxUINT8_CLASS,
}

# Before release v7.1 (release 14) matlab (TM) used the system
# default character encoding scheme padded out to 16-bits. Release 14
# and later use Unicode. When saving character data, R14 checks if it
# can be encoded in 7-bit ascii, and saves in that format if so.

codecs_template = {
    miTypes.miUTF8: {"codec": "utf_8", "width": 1},
    miTypes.miUTF16: {"codec": "utf_16", "width": 2},
    miTypes.miUTF32: {"codec": "utf_32", "width": 4},
}


def _convert_codecs(template, byte_order):
    """Convert codec template mapping to byte order

    Set codecs not on this system to None

    Parameters
    ----------
    template : mapping
       key, value are respectively codec name, and root name for codec
       (without byte order suffix)
    byte_order : {'<', '>'}
       code for little or big endian

    Returns
    -------
    codecs : dict
       key, value are name, codec (as in .encode(codec))
    """
    codecs = {}
    postfix = byte_order == "<" and "_le" or "_be"
    for k, v in template.items():
        codec = v["codec"]
        try:
            " ".encode(codec)
        except LookupError:
            codecs[k] = None
            continue
        if v["width"] > 1:
            codec += postfix
        codecs[k] = codec
    return codecs.copy()


def convert_dtypes(dtype_template, order_code):
    """Convert dtypes in mapping to given order

    Parameters
    ----------
    dtype_template : mapping
       mapping with values returning numpy dtype from ``np.dtype(val)``
    order_code : str
       an order code suitable for using in ``dtype.newbyteorder()``

    Returns
    -------
    dtypes : mapping
       mapping where values have been replaced by
       ``np.dtype(val).newbyteorder(order_code)``

    """
    dtypes = dtype_template.copy()
    for k in dtypes:
        dtypes[k] = np.dtype(dtypes[k]).newbyteorder(order_code)
    return dtypes


MDTYPES = {}
for _bytecode in "<>":
    _def = {
        "dtypes": convert_dtypes(mdtypes_template, _bytecode),
        "classes": convert_dtypes(mclass_dtypes_template, _bytecode),
        "codecs": _convert_codecs(codecs_template, _bytecode),
    }
    MDTYPES[_bytecode] = _def



from .matclass import (
    MatlabContainerMap,
    MatlabEnumerationArray,
    MatlabFunction,
    MatlabObject,
    MatlabOpaque,
    MatlabOpaqueArray,
)

__all__ = [
    "MatlabContainerMap",
    "MatlabEnumerationArray",
    "MatlabOpaque",
    "MatlabOpaqueArray",
    "MatlabFunction",
    "MatlabObject",
]



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat

files = [("test_basic_v7.mat", "v7"), ("test_basic_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadSparse:
    """Tests for reading sparse matrices from MAT-files"""

    def test_sparse_empty(self, filename, version):
        """Test reading Sparse Empty from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_empty"], spmatrix=False
        )
        assert "sparse_empty" in mdict

        sp = np.empty((0, 0), dtype=np.float64)
        np.testing.assert_array_equal(mdict["sparse_empty"].toarray(), sp, strict=True)

    def test_sparse_col(self, filename, version):
        """Test reading Sparse Column from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_col"], spmatrix=False)
        assert "sparse_col" in mdict

        sp = np.array([[0], [1], [0], [3]], dtype=np.float64)
        np.testing.assert_array_equal(mdict["sparse_col"].toarray(), sp, strict=True)

    def test_sparse_row(self, filename, version):
        """Test reading Sparse Row from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_row"], spmatrix=False)
        assert "sparse_row" in mdict

        sp = np.array([[0, 5, 0, 0]], dtype=np.float64)
        np.testing.assert_array_equal(mdict["sparse_row"].toarray(), sp, strict=True)

    def test_sparse_diag(self, filename, version):
        """Test reading Sparse Diagonal from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_diag"], spmatrix=False)
        assert "sparse_diag" in mdict

        sp = np.array(
            [
                [1, 0, 0, 0, 0],
                [0, 2, 0, 0, 0],
                [0, 0, 3, 0, 0],
                [0, 0, 0, 4, 0],
                [0, 0, 0, 0, 5],
            ],
            dtype=np.float64,
        )
        np.testing.assert_array_equal(mdict["sparse_diag"].toarray(), sp, strict=True)

    def test_sparse_rec_row(self, filename, version):
        """Test reading Sparse Rectangular Row from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_rec_row"], spmatrix=False
        )
        assert "sparse_rec_row" in mdict

        sp = np.array([[1, 0], [0, 2], [3, 0], [0, 4]], dtype=np.float64)
        np.testing.assert_array_equal(
            mdict["sparse_rec_row"].toarray(), sp, strict=True
        )

    def test_sparse_rec_col(self, filename, version):
        """Test reading Sparse Rectangular Column from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_rec_col"], spmatrix=False
        )
        assert "sparse_rec_col" in mdict

        sp = np.array([[1, 0, 0, 2], [0, 3, 0, 0]], dtype=np.float64)
        np.testing.assert_array_equal(
            mdict["sparse_rec_col"].toarray(), sp, strict=True
        )

    def test_sparse_symmetric(self, filename, version):
        """Test reading Sparse Symmetric from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_symmetric"], spmatrix=False
        )
        assert "sparse_symmetric" in mdict

        sp = np.array([[1, 2, 0], [2, 3, 4], [0, 4, 5]], dtype=np.float64)
        np.testing.assert_array_equal(
            mdict["sparse_symmetric"].toarray(), sp, strict=True
        )

    def test_sparse_neg(self, filename, version):
        """Test reading Sparse with Negative Values from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_neg"], spmatrix=False)
        assert "sparse_neg" in mdict

        sp = np.array([[0, -1, 0], [2, 0, 0], [0, 0, 3]], dtype=np.float64)
        np.testing.assert_array_equal(mdict["sparse_neg"].toarray(), sp, strict=True)

    def test_sparse_logical(self, filename, version):
        """Test reading Sparse Logical from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_logical"], spmatrix=False
        )
        assert "sparse_logical" in mdict

        sp = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]], dtype=np.bool_)
        np.testing.assert_array_equal(
            mdict["sparse_logical"].toarray(), sp, strict=True
        )

    def test_sparse_complex(self, filename, version):
        """Test reading Sparse Complex from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_complex"], spmatrix=False
        )
        assert "sparse_complex" in mdict

        sp = np.array(
            [[1 + 1j, 0, 0], [0, 2 - 2j, 0], [0, 0, 3 + 3j]], dtype=np.complex128
        )
        np.testing.assert_array_equal(
            mdict["sparse_complex"].toarray(), sp, strict=True
        )

    def test_sparse_nnz(self, filename, version):
        """Test reading sparse with all non-zero elements from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_nnz"], spmatrix=False)
        assert "sparse_nnz" in mdict

        sp = np.array([[1, 2], [3, 4]], dtype=np.float64)
        np.testing.assert_array_equal(mdict["sparse_nnz"].toarray(), sp, strict=True)

    def test_sparse_all_zeros(self, filename, version):
        """Test reading sparse with all zero elements from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(
            file_path, variable_names=["sparse_all_zeros"], spmatrix=False
        )
        assert "sparse_all_zeros" in mdict

        sp = np.array([[0, 0], [0, 0]], dtype=np.float64)
        np.testing.assert_array_equal(
            mdict["sparse_all_zeros"].toarray(), sp, strict=True
        )


@pytest.mark.parametrize("filename, version", files)
class TestSaveSparse:

    def test_sparse_empty(self, filename, version):
        """Test writing empty sparse to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_empty"])

            np.testing.assert_array_equal(
                mdict["sparse_empty"].toarray(),
                mload["sparse_empty"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_col(self, filename, version):
        """Test writing sparse column to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_col"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_col"])

            np.testing.assert_array_equal(
                mdict["sparse_col"].toarray(),
                mload["sparse_col"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_row(self, filename, version):
        """Test writing sparse row to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_row"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_row"])

            np.testing.assert_array_equal(
                mdict["sparse_row"].toarray(),
                mload["sparse_row"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_diag(self, filename, version):
        """Test writing sparse diagonal matrix to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_diag"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_diag"])

            np.testing.assert_array_equal(
                mdict["sparse_diag"].toarray(),
                mload["sparse_diag"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_rec_row(self, filename, version):
        """Test writing sparse rectangular row to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_rec_row"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_rec_row"])

            np.testing.assert_array_equal(
                mdict["sparse_rec_row"].toarray(),
                mload["sparse_rec_row"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_rec_col(self, filename, version):
        """Test writing sparse rectangular column to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_rec_col"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_rec_col"])

            np.testing.assert_array_equal(
                mdict["sparse_rec_col"].toarray(),
                mload["sparse_rec_col"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_symmetric(self, filename, version):
        """Test writing sparse symmetric to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_symmetric"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_symmetric"])

            np.testing.assert_array_equal(
                mdict["sparse_symmetric"].toarray(),
                mload["sparse_symmetric"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_neg(self, filename, version):
        """Test writing sparse matrix with negative values to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_neg"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_neg"])

            np.testing.assert_array_equal(
                mdict["sparse_neg"].toarray(),
                mload["sparse_neg"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_logical(self, filename, version):
        """Test writing sparse matrix with logical values to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_logical"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_logical"])

            np.testing.assert_array_equal(
                mdict["sparse_logical"].toarray(),
                mload["sparse_logical"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_complex(self, filename, version):
        """Test writing sparse matrix with complex values to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_complex"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_complex"])

            np.testing.assert_array_equal(
                mdict["sparse_complex"].toarray(),
                mload["sparse_complex"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_sparse_nnz(self, filename, version):
        """Test writing sparse with all non-zeros to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["sparse_nnz"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["sparse_nnz"])

            np.testing.assert_array_equal(
                mdict["sparse_nnz"].toarray(),
                mload["sparse_nnz"].toarray(),
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)



import numpy
from Cython.Build import cythonize
from setuptools import Extension, setup

extensions = [
    Extension(
        "matio.v5._streams",
        ["matio/v5/_streams.pyx"],
        include_dirs=[numpy.get_include()],
    ),
    Extension(
        "matio.v5._mio5_utils",
        ["matio/v5/_mio5_utils.pyx"],
        include_dirs=[numpy.get_include(), "src/v5/include"],
    ),
]

setup(
    ext_modules=cythonize(extensions, include_path=["matio"]),
)



import os

import pytest

from matio import load_from_mat
from matio.utils.matclass import MatlabOpaque, MatReadWarning

files = [("test_type_systems_v7.mat", "v7"), ("test_type_systems_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadTypeSystems:

    def test_type_system(self, filename, version):
        """Test reading different type systems from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        with pytest.warns(MatReadWarning):
            mdict = load_from_mat(file_path, variable_names=["javatype", "handletype"])

        assert "javatype" in mdict
        assert "handletype" in mdict

        assert isinstance(mdict["javatype"], MatlabOpaque)
        assert isinstance(mdict["handletype"], MatlabOpaque)

        assert mdict["javatype"].classname == "java.lang.String"
        assert mdict["handletype"].classname == "COM.Excel_Application"
        assert mdict["javatype"].type_system == "java"
        assert mdict["handletype"].type_system == "handle"



import struct
import sys
import time
from enum import StrEnum

import h5py
import numpy as np

from matio.utils.matclass import MatReadError

MAT5_HEADER_SIZE_BYTES = 128
MAT5_MAX_ARR_BYTES = 2**32
MAT5_MAX_STRUCT_FIELDNAME_LEN = 64
MAT_5_VERSION = 1

MAT_HDF_VERSION = 2
MAT_HDF_USER_BLOCK_BYTES = 512
MAT_HDF_REFS_GROUP = "#refs#"
MAT_HDF_SUBSYS_GROUP = "#subsystem#"
MAT_HDF_COMPRESSION = "gzip"
MAT_HDF_COMPRESSION_OPTS = 3

MCOS_MAGIC_NUMBER = 0xDD000000

MAT_VERSIONS = {"v7": MAT_5_VERSION, "v7.3": MAT_HDF_VERSION}


class MAT_HDF_ATTRS(StrEnum):
    """Enumeration for standard HDF5 attributes in MAT v7.3 files"""

    CLASS = "MATLAB_class"
    OBJECT_DECODE = "MATLAB_object_decode"
    INT_DECODE = "MATLAB_int_decode"
    EMPTY = "MATLAB_empty"
    GLOBAL = "MATLAB_global"
    SPARSE = "MATLAB_sparse"
    FIELDS = "MATLAB_fields"


def check_mat_version(data):
    """Reads MAT-file version from header data"""

    if data[2:] == b"IM":
        byte_order = "<"
    elif data[2:] == b"MI":
        byte_order = ">"
    else:
        raise MatReadError("Invalid endian indicator in MAT-file header")

    v_major, v_minor = int(data[1]), int(data[0])
    if byte_order != "<":
        v_major, v_minor = v_minor, v_major

    if v_major not in (MAT_5_VERSION, MAT_HDF_VERSION):
        raise MatReadError(f"Unknown MAT-file version {v_major}.{v_minor}")

    return byte_order, v_major


def read_mat_header(file_path):
    """Reads MAT-file header and returns version information"""

    with open(file_path, "rb") as f:
        data = f.read(MAT5_HEADER_SIZE_BYTES)
        byte_order, v_major = check_mat_version(data[124:])

        subsystem_offset = np.frombuffer(data[116:124], dtype=byte_order + "u8")[0]
        if subsystem_offset == 0x2020202020202020:
            subsystem_offset = 0  # All spaces in BE Files

        return subsystem_offset, v_major, byte_order


def write_subsystem_offset(file_stream, offset=0):
    """Write 8 bytes of subsystem offset at byte 116"""

    file_stream.seek(116)
    file_stream.write(struct.pack("<Q", offset))


def write_version(file_stream, version):
    """Write version information"""

    v_major = version
    v_minor = 0

    is_little_endian = sys.byteorder == "little"

    if is_little_endian:
        file_stream.write(struct.pack("<BB", v_minor, v_major))
        file_stream.write(b"IM")
    else:
        file_stream.write(struct.pack(">BB", v_major, v_minor))
        file_stream.write(b"MI")


def write_file_header(file_stream, version):
    """Write MAT-file header"""

    current_time = time.strftime("%Y-%m-%d %H:%M:%S UTC", time.gmtime())
    description = (
        f"MATLAB 5.0 MAT-file Platform. " f"Created on: {current_time} by matio"
    )
    if version == MAT_5_VERSION:
        description += " with scipy"
    elif version == MAT_HDF_VERSION:
        description += f" using h5py v{h5py.__version__}"

    description_bytes = description.encode("ascii")[:116]  # Truncate if too long
    description_padded = description_bytes.ljust(116, b"\x20")

    file_stream.write(description_padded)
    write_subsystem_offset(file_stream)
    write_version(file_stream, version)



"""Utility functions for convertin MATLAB strings"""

import warnings

import numpy as np

from matio.utils.matclass import MatConvertWarning

MAT_STRING_VERSION = 1


def mat_to_string(props, byte_order, **_kwargs):
    """Converts MATLAB string to numpy string array"""

    data = props.get("any", np.empty((0, 0), dtype=np.str_))
    if data.size == 0:
        return np.array([[]], dtype=np.dtypes.StringDType())

    if data[0, 0] != MAT_STRING_VERSION:
        warnings.warn(
            "mat_to_string: String saved from a different MAT-file version. Returning raw data",
            MatConvertWarning,
        )
        return props[0, 0].get("any")

    ndims = data[0, 1]
    shape = data[0, 2 : 2 + ndims]
    num_strings = np.prod(shape)
    char_counts = data[0, 2 + ndims : 2 + ndims + num_strings]
    byte_data = data[0, 2 + ndims + num_strings :].tobytes()

    strings = []
    pos = 0
    encoding = "utf-16-le" if byte_order[0] == "<" else "utf-16-be"
    for char_count in char_counts:
        if char_count == 0xFFFFFFFFFFFFFFFF:
            strings.append(np.nan)
            print("here")
            continue
        byte_length = char_count * 2  # UTF-16 encoding
        extracted_string = byte_data[pos : pos + byte_length].decode(encoding)
        strings.append(np.str_(extracted_string))
        pos += byte_length

    arr = np.array(strings, dtype=np.dtypes.StringDType(na_object=np.nan)).reshape(
        shape, order="F"
    )
    return arr


def string_to_mat(arr):
    """Converts numpy string array to MATLAB string format as uint64 array"""

    ndims = arr.ndim
    shape = arr.shape
    encoding = "utf-16-le" if np.little_endian else "utf-16-be"

    utf16_data_list = []
    char_counts = []

    for s in arr.ravel(order="F"):
        utf16_arr = np.frombuffer(s.encode(encoding), dtype=np.uint16)
        utf16_data_list.append(utf16_arr)
        char_counts.append(len(utf16_arr))
    all_utf16 = np.hstack(utf16_data_list)

    # Pad
    pad_len = (-len(all_utf16)) % 4
    if pad_len > 0:
        all_utf16 = np.hstack([all_utf16, np.zeros(pad_len, dtype=np.uint16)])
    utf16_data_uint64 = all_utf16.view(np.uint64)

    # Header
    header_list = [MAT_STRING_VERSION, ndims] + list(shape) + char_counts
    header = np.array(header_list, dtype=np.uint64)
    prop = np.hstack([header, utf16_data_uint64]).reshape(1, -1)

    prop_map = {
        "any": prop,
    }

    return prop_map



import os
import tempfile

import numpy as np
import pytest

from matio import load_from_mat, save_to_mat

files = [("test_time_v7.mat", "v7"), ("test_time_v73.mat", "v7.3")]


@pytest.mark.parametrize("filename, version", files)
class TestLoadMatlabCalendarDuration:

    def test_calendarDuration_days(self, filename, version):
        """Test reading calendarDuration days scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days"])
        assert "cdur_days" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_days = np.array(
            [
                (0, 1, 0),
                (0, 2, 0),
                (0, 3, 0),
            ],
            dtype=out_dtype,
        ).reshape(1, 3)

        np.testing.assert_array_equal(mdict["cdur_days"], cdur_days, strict=True)

    def test_calendarDuration_weeks(self, filename, version):
        """Test reading calendarDuration weeks scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_weeks"])
        assert "cdur_weeks" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_weeks = np.array(
            [
                (0, 7, 0),
                (0, 14, 0),
            ],
            dtype=out_dtype,
        ).reshape(1, 2)

        np.testing.assert_array_equal(mdict["cdur_weeks"], cdur_weeks, strict=True)

    def test_calendarDuration_days_and_months(self, filename, version):
        """Test reading calendarDuration days scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days_and_months"])
        assert "cdur_days_and_months" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_days_and_weeks = np.array([(1, 1, 0), (0, 2, 0)], dtype=out_dtype).reshape(
            1, 2
        )

        np.testing.assert_array_equal(
            mdict["cdur_days_and_months"], cdur_days_and_weeks, strict=True
        )

    def test_calendarDuration_months_and_years(self, filename, version):
        """Test reading calendarDuration months scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_months_and_years"])
        assert "cdur_months_and_years" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_months_and_years = np.array(
            [(12, 0, 0), (18, 0, 0)], dtype=out_dtype
        ).reshape(1, 2)

        np.testing.assert_array_equal(
            mdict["cdur_months_and_years"], cdur_months_and_years, strict=True
        )

    def test_calendarDuration_days_and_quarters(self, filename, version):
        """Test reading calendarDuration days and months scalar data from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days_and_qtrs"])
        assert "cdur_days_and_qtrs" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_days_and_quarters = np.array(
            [
                (3, 15, 0),
            ],
            dtype=out_dtype,
        ).reshape(1, 1)

        np.testing.assert_array_equal(
            mdict["cdur_days_and_qtrs"], cdur_days_and_quarters, strict=True
        )

    def test_calendarDuration_array(self, filename, version):
        """Test reading calendarDuration array from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_array"])
        assert "cdur_array" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_array = np.array(
            [
                (1, 0, 0),
                (0, 5, 0),
                (2, 0, 0),
                (0, 10, 0),
            ],
            dtype=out_dtype,
        ).reshape(2, 2)

        np.testing.assert_array_equal(mdict["cdur_array"], cdur_array, strict=True)

    def test_calendarDuration_millis(self, filename, version):
        """Test reading calendarDuration millis scalar from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_millis"])
        assert "cdur_millis" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_millis = np.array(
            [
                (0, 1, 3723000),
            ],
            dtype=out_dtype,
        ).reshape(1, 1)

        np.testing.assert_array_equal(mdict["cdur_millis"], cdur_millis, strict=True)

    def test_calendarDuration_empty(self, filename, version):
        """Test reading calendarDuration empty from MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_empty"])
        assert "cdur_empty" in mdict

        out_dtype = [
            ("months", "timedelta64[M]"),
            ("days", "timedelta64[D]"),
            ("millis", "timedelta64[ms]"),
        ]

        cdur_empty = np.empty((0, 0), dtype=out_dtype)

        np.testing.assert_array_equal(mdict["cdur_empty"], cdur_empty, strict=True)


@pytest.mark.parametrize("filename, version", files)
class TestSaveMatlabCalendarDuration:

    def test_calendarDuration_days(self, filename, version):
        """Test writing calendarDuration days scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_days"])

            np.testing.assert_array_equal(
                mdict["cdur_days"], mload["cdur_days"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_weeks(self, filename, version):
        """Test writing calendarDuration weeks scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_weeks"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_weeks"])

            np.testing.assert_array_equal(
                mdict["cdur_weeks"], mload["cdur_weeks"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_days_and_months(self, filename, version):
        """Test writing calendarDuration days and months scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days_and_months"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["cdur_days_and_months"]
            )

            np.testing.assert_array_equal(
                mdict["cdur_days_and_months"],
                mload["cdur_days_and_months"],
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_months_and_years(self, filename, version):
        """Test writing calendarDuration months scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_months_and_years"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(
                temp_file_path, variable_names=["cdur_months_and_years"]
            )

            np.testing.assert_array_equal(
                mdict["cdur_months_and_years"],
                mload["cdur_months_and_years"],
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_days_and_quarters(self, filename, version):
        """Test writing calendarDuration months scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_days_and_qtrs"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_days_and_qtrs"])

            np.testing.assert_array_equal(
                mdict["cdur_days_and_qtrs"],
                mload["cdur_days_and_qtrs"],
                strict=True,
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_array(self, filename, version):
        """Test writing calendarDuration array to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_array"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_array"])

            np.testing.assert_array_equal(
                mdict["cdur_array"], mload["cdur_array"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_millis(self, filename, version):
        """Test writing calendarDuration millis scalar to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_millis"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_millis"])

            np.testing.assert_array_equal(
                mdict["cdur_millis"], mload["cdur_millis"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)

    def test_calendarDuration_empty(self, filename, version):
        """Test writing calendarDuration empty to MAT-file"""
        file_path = os.path.join(os.path.dirname(__file__), "data", filename)
        mdict = load_from_mat(file_path, variable_names=["cdur_empty"])

        with tempfile.NamedTemporaryFile(suffix=".mat", delete=False) as tmpfile:
            temp_file_path = tmpfile.name

        try:
            save_to_mat(temp_file_path, mdict, version=version)
            mload = load_from_mat(temp_file_path, variable_names=["cdur_empty"])

            np.testing.assert_array_equal(
                mdict["cdur_empty"], mload["cdur_empty"], strict=True
            )

        finally:
            if os.path.exists(temp_file_path):
                os.remove(temp_file_path)
